@@ -27,7 +27,7 @@ Example:
 
 from ctypes import cdll, c_void_p, c_int, c_char_p, c_uint, byref
 from enum import IntEnum
-from sys import platform
+import sys
 
 CTL_ALLOC = None
 CTL_FREE = None
