@@ -24,7 +24,7 @@
 #include "knot/ctl/commands.h"
 #include "knot/dnssec/key-events.h"
 #include "knot/events/events.h"
-#include "knot/events/log.h"
+#include "knot/events/handlers.h"
 #include "knot/nameserver/query_module.h"
 #include "knot/updates/zone-update.h"
 #include "knot/zone/timers.h"
@@ -363,7 +363,7 @@ static int zone_ksk_sbm_confirm(zone_t *zone, ctl_args_t *args)
 		return ret;
 	}
 
-	ret = knot_dnssec_ksk_sbm_confirm(&ctx);
+	ret = knot_dnssec_ksk_sbm_confirm(&ctx, 0);
 	kdnssec_ctx_deinit(&ctx);
 
 	conf_val_t val = conf_zone_get(conf(), C_DNSSEC_SIGNING, zone->name);
@@ -453,8 +453,7 @@ static int zone_txn_commit(zone_t *zone, ctl_args_t *args)
 			zone_txn_update_clear(zone);
 			return ret;
 		}
-		log_dnssec_next(zone->name, (time_t)resch.next_sign);
-		zone_events_schedule_at(zone, ZONE_EVENT_DNSSEC, (time_t)resch.next_sign);
+		event_dnssec_reschedule(conf(), zone, &resch, false);
 	}
 
 	int ret = zone_update_commit(conf(), zone->control_update);
