@@ -313,6 +313,7 @@ static knot_time_t ksk_retire_time(knot_time_t retire_active_time, const kdnssec
 	if (retire_active_time <= 0) {
 		return 0;
 	}
+	// this is not correct! It should be parent DS TTL.
 	return knot_time_add(retire_active_time, ctx->policy->propagation_delay + ctx->policy->dnskey_ttl);
 }
 
@@ -377,8 +378,11 @@ static roll_action_t next_action(kdnssec_ctx_t *ctx)
 				}
 				break;
 			case DNSSEC_KEY_STATE_RETIRE_ACTIVE:
-				keytime = ksk_retire_time(key->timing.retire_active, ctx);
-				restype = RETIRE;
+				if (key->timing.retire == 0 && key->timing.post_active == 0) { // this shouldn't normally happen
+					// when a KSK is retire_active, it has already retire or post_active timer set
+					keytime = ksk_retire_time(key->timing.retire_active, ctx);
+					restype = RETIRE;
+				}
 				break;
 			case DNSSEC_KEY_STATE_POST_ACTIVE:
 				keytime = alg_remove_time(key->timing.post_active, ctx);
@@ -462,7 +466,7 @@ static int submit_key(kdnssec_ctx_t *ctx, knot_kasp_key_t *newkey)
 	return KNOT_EOK;
 }
 
-static int exec_new_signatures(kdnssec_ctx_t *ctx, knot_kasp_key_t *newkey)
+static int exec_new_signatures(kdnssec_ctx_t *ctx, knot_kasp_key_t *newkey, uint32_t active_retire_delay)
 {
 	if (newkey->is_ksk) {
 		log_zone_notice(ctx->zone->dname, "DNSSEC, KSK submission, confirmed");
@@ -480,6 +484,14 @@ static int exec_new_signatures(kdnssec_ctx_t *ctx, knot_kasp_key_t *newkey)
 				key->timing.retire = ctx->now;
 			}
 		}
+		if (newkey->is_ksk && (keystate == DNSSEC_KEY_STATE_ACTIVE ||
+		                       keystate == DNSSEC_KEY_STATE_RETIRE_ACTIVE)) {
+			if (keyalg != dnssec_key_get_algorithm(newkey->key)) {
+				key->timing.post_active = ctx->now + active_retire_delay;
+			} else if (key->is_ksk) {
+				key->timing.retire = ctx->now + active_retire_delay;
+			}
+		}
 	}
 
 	if (newkey->is_ksk) {
@@ -623,7 +635,7 @@ int knot_dnssec_key_rollover(kdnssec_ctx_t *ctx, zone_sign_reschedule_t *resched
 			reschedule->plan_ds_query = true;
 			break;
 		case REPLACE:
-			ret = exec_new_signatures(ctx, next.key);
+			ret = exec_new_signatures(ctx, next.key, 0);
 			break;
 		case RETIRE:
 			ret = exec_ksk_retire(ctx, next.key);
@@ -657,12 +669,12 @@ int knot_dnssec_key_rollover(kdnssec_ctx_t *ctx, zone_sign_reschedule_t *resched
 	return ret;
 }
 
-int knot_dnssec_ksk_sbm_confirm(kdnssec_ctx_t *ctx)
+int knot_dnssec_ksk_sbm_confirm(kdnssec_ctx_t *ctx, uint32_t retire_delay)
 {
 	for (size_t i = 0; i < ctx->zone->num_keys; i++) {
 		knot_kasp_key_t *key = &ctx->zone->keys[i];
 		if (key->is_ksk && get_key_state(key, ctx->now) == DNSSEC_KEY_STATE_READY) {
-			int ret = exec_new_signatures(ctx, key);
+			int ret = exec_new_signatures(ctx, key, retire_delay);
 			if (ret == KNOT_EOK) {
 				ret = kdnssec_ctx_commit(ctx);
 			}
