@@ -1,4 +1,4 @@
-/*  Copyright (C) 2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -17,8 +17,8 @@
 #pragma once
 
 #include "contrib/dynarray.h"
-#include "dnssec/keystore.h"
-#include "dnssec/sign.h"
+#include "libdnssec/keystore.h"
+#include "libdnssec/sign.h"
 #include "knot/dnssec/kasp/kasp_zone.h"
 #include "knot/dnssec/kasp/policy.h"
 #include "knot/dnssec/context.h"
@@ -52,21 +52,36 @@ typedef struct {
 /*!
  * \brief Flags determining key type
  */
-extern const uint16_t DNSKEY_FLAGS_KSK;
-extern const uint16_t DNSKEY_FLAGS_ZSK;
-uint16_t dnskey_flags(bool is_ksk);
+enum {
+	DNSKEY_FLAGS_KSK = KNOT_DNSKEY_FLAG_ZONE | KNOT_DNSKEY_FLAG_SEP,
+	DNSKEY_FLAGS_ZSK = KNOT_DNSKEY_FLAG_ZONE,
+};
+
+inline static uint16_t dnskey_flags(bool is_ksk)
+{
+	return is_ksk ? DNSKEY_FLAGS_KSK : DNSKEY_FLAGS_ZSK;
+}
+
+typedef enum {
+	DNSKEY_GENERATE_KSK      = (1 << 0), // KSK flag in metadata
+	DNSKEY_GENERATE_ZSK      = (1 << 1), // ZSK flag in metadata
+	DNSKEY_GENERATE_SEP_SPEC = (1 << 2), // not (SEP bit set iff KSK)
+	DNSKEY_GENERATE_SEP_ON   = (1 << 3), // SEP bit set on
+} kdnssec_generate_flags_t;
+
+void normalize_generate_flags(kdnssec_generate_flags_t *flags);
 
 /*!
  * \brief Generate new key, store all details in new kasp key structure.
  *
  * \param ctx           kasp context
- * \param ksk           true = generate KSK or CSK, false = generate ZSK
- * \param zsk           true = generate ZSK or CSK, false = generate KSK
+ * \param flags         determine if to use the key as KSK and/or ZSK and SEP flag
  * \param key_ptr       output if KNOT_EOK: new pointer to generated key
  *
  * \return KNOT_E*
  */
-int kdnssec_generate_key(kdnssec_ctx_t *ctx, bool ksk, bool zsk, knot_kasp_key_t **key_ptr);
+int kdnssec_generate_key(kdnssec_ctx_t *ctx, kdnssec_generate_flags_t flags,
+                         knot_kasp_key_t **key_ptr);
 
 /*!
  * \brief Take a key from another zone (copying info, sharing privkey).
