@@ -1,4 +1,4 @@
-/*  Copyright (C) 2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -14,6 +14,7 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+#include <inttypes.h>
 #include <limits.h>
 #include <sys/stat.h>
 #include <stdarg.h>
@@ -21,7 +22,9 @@
 #include "knot/journal/journal.h"
 #include "knot/common/log.h"
 #include "contrib/files.h"
-#include "contrib/endian.h"
+#include "contrib/ctype.h"
+#include "libknot/endian.h"
+#include "contrib/dynarray.h"
 
 /*! \brief Journal version. */
 #define JOURNAL_VERSION	"1.0"
@@ -117,7 +120,7 @@ typedef struct {
 	uint32_t flags;           // LAST_FLUSHED_VALID, SERIAL_TO_VALID, MERGED_SERIAL_VALID.
 } metadata_t;
 
-typedef struct {
+typedef struct journal_txn {
 	journal_t *j;
 	knot_db_txn_t *txn;
 	int ret;
@@ -168,10 +171,25 @@ static void txn_init(txn_t *txn, knot_db_txn_t *db_txn, journal_t *j)
  *
  */
 
+static bool key_is_ok(const knot_db_val_t *key, bool zone_related)
+{
+	const uint8_t *it = key->data;
+	ssize_t it_len = key->len;
+	if (zone_related) {
+		size_t dname_len = knot_dname_size(it);
+		it += dname_len;
+		it_len -= dname_len;
+	}
+	it += 4;
+	it_len -= 4;
+
+	return ((zone_related && it_len == 8) || // normal changeset
+	       (is_lower(*it) && !is_lower(*(it-1)))); // metadata
+}
+
 static void txn_key_str(txn_t *txn, const knot_dname_t *zone, const char *key)
 {
-	size_t zone_size = 0;
-	if (zone != NULL) zone_size = knot_dname_size(zone);
+	size_t zone_size = knot_dname_size(zone);
 	txn->key.len = zone_size + DB_KEY_UNUSED_ZERO + strlen(key) + 1;
 	if (txn->key.len > 512) {
 		txn->ret = KNOT_ERROR;
@@ -180,12 +198,12 @@ static void txn_key_str(txn_t *txn, const knot_dname_t *zone, const char *key)
 	if (zone != NULL) memcpy(txn->key.data, zone, zone_size);
 	memset(txn->key.data + zone_size, 0, DB_KEY_UNUSED_ZERO);
 	strcpy(txn->key.data + zone_size + DB_KEY_UNUSED_ZERO, key);
+	assert(key_is_ok(&txn->key, zone != NULL));
 }
 
 static void txn_key_2u32(txn_t *txn, const knot_dname_t *zone, uint32_t key1, uint32_t key2)
 {
-	size_t zone_size = 0;
-	if (zone != NULL) zone_size = knot_dname_size(zone);
+	size_t zone_size = knot_dname_size(zone);
 	txn->key.len = zone_size + DB_KEY_UNUSED_ZERO + 2*sizeof(uint32_t);
 	if (txn->key.len > 512) {
 		txn->ret = KNOT_ERROR;
@@ -198,12 +216,12 @@ static void txn_key_2u32(txn_t *txn, const knot_dname_t *zone, uint32_t key1, ui
 	memcpy(txn->key.data + zone_size + DB_KEY_UNUSED_ZERO, &key_be1, sizeof(uint32_t));
 	memcpy(txn->key.data + zone_size + DB_KEY_UNUSED_ZERO + sizeof(uint32_t),
 	       &key_be2, sizeof(uint32_t));
+	assert(key_is_ok(&txn->key, zone != NULL));
 }
 
 static void txn_key_str_u32(txn_t *txn, const knot_dname_t *zone, const char *key1, uint32_t key2)
 {
-	size_t zone_size = 0;
-	if (zone != NULL) zone_size = knot_dname_size(zone);
+	size_t zone_size = knot_dname_size(zone);
 	txn->key.len = zone_size + DB_KEY_UNUSED_ZERO + strlen(key1) + 1 + sizeof(uint32_t);
 	if (txn->key.len > 512) {
 		txn->ret = KNOT_ERROR;
@@ -215,6 +233,7 @@ static void txn_key_str_u32(txn_t *txn, const knot_dname_t *zone, const char *ke
 	uint32_t key_be2 = htobe32(key2);
 	memcpy(txn->key.data + zone_size + DB_KEY_UNUSED_ZERO + strlen(key1) + 1,
 	       &key_be2, sizeof(uint32_t));
+	assert(key_is_ok(&txn->key, zone != NULL));
 }
 
 static int txn_cmpkey(txn_t *txn, knot_db_val_t *key2)
@@ -397,6 +416,13 @@ static void txn_commit(txn_t *txn)
 	txn_abort(txn); // no effect if all ok
 }
 
+void journal_txn_commit(struct journal_txn *txn)
+{
+	if (txn != NULL) {
+		txn_commit(txn);
+	}
+}
+
 static void txn_restart(txn_t *txn)
 {
 	txn_commit(txn);
@@ -410,8 +436,7 @@ static void txn_reuse(txn_t **txn, txn_t *to_reuse, bool write_allowed)
 {
 	if (to_reuse == NULL) {
 		txn_begin(*txn, write_allowed);
-	}
-	else {
+	} else {
 		*txn = to_reuse;
 	}
 }
@@ -463,8 +488,7 @@ static void md_get_common_last_inserter_zone(txn_t *txn, knot_dname_t **res)
 	txn_key_str(txn, NULL, MDKEY_GLOBAL_LAST_INSERTER_ZONE);
 	if (txn_find(txn)) {
 		*res = knot_dname_copy(txn->val.data, NULL);
-	}
-	else {
+	} else {
 		*res = NULL;
 	}
 }
@@ -597,8 +621,7 @@ static int initial_md_check(journal_t *j, bool *dirty_present)
 		txn->val.data = JOURNAL_VERSION;
 		txn_insert(txn);
 		something_updated = true;
-	}
-	else {
+	} else {
 		char * jver = txn->val.data;
 		if (first_digit(jver) != first_digit(JOURNAL_VERSION)) {
 			txn_abort(txn);
@@ -614,8 +637,7 @@ static int initial_md_check(journal_t *j, bool *dirty_present)
 
 	if (something_updated) {
 		txn_commit(txn);
-	}
-	else { // abort to gain up speed when opening a lot of zones
+	} else { // abort to gain up speed when opening a lot of zones
 		txn_abort(txn);
 	}
 
@@ -733,8 +755,7 @@ static int iterate(journal_t *j, txn_t *_txn, iteration_cb_t cb, int method,
 
 			ctx.serial = ctx.serial_to;
 			ctx.chunk_index = 0;
-		}
-		else {
+		} else {
 			ctx.chunk_index++;
 		}
 
@@ -801,6 +822,30 @@ static int vals_to_changeset(knot_db_val_t *vals, int nvals,
 	return KNOT_EOK;
 }
 
+static int vals_to_chgset_ctx(knot_db_val_t *vals, int nvals, uint32_t serial_from,
+                              uint32_t serial_to, chgset_ctx_t **ch)
+{
+	if (nvals < 1) {
+		return KNOT_EINVAL;
+	}
+
+	chgset_ctx_t *t_ch = chgset_ctx_create(nvals);
+	if (t_ch == NULL) {
+		return KNOT_ENOMEM;
+	}
+
+	for (size_t i = 0; i < nvals; i++) {
+		t_ch->src_chunks[i] = vals[i].data + JOURNAL_HEADER_SIZE;
+		t_ch->chunk_sizes[i] = vals[i].len - JOURNAL_HEADER_SIZE;
+	}
+
+	t_ch->serial_from = serial_from;
+	t_ch->serial_to = serial_to;
+
+	*ch = t_ch;
+	return KNOT_EOK;
+}
+
 static int load_one_itercb(iteration_ctx_t *ctx)
 {
 	changeset_t *ch = NULL, **targ = ctx->iter_context;
@@ -826,6 +871,19 @@ static int load_list_itercb(iteration_ctx_t *ctx)
 	return ret;
 }
 
+static int load_list_ctx_itercb(iteration_ctx_t *ctx)
+{
+	chgset_ctx_t *ch = NULL;
+	list_t *chlist = *(list_t **) ctx->iter_context;
+
+	int ret = vals_to_chgset_ctx(ctx->val, ctx->chunk_count, ctx->serial, ctx->serial_to, &ch);
+
+	if (ret == KNOT_EOK) {
+		add_tail(chlist, &ch->n);
+	}
+	return ret;
+}
+
 /*! \brief Load one changeset (with serial) from DB */
 static int load_one(journal_t *j, txn_t *_txn, uint32_t serial, changeset_t **ch)
 {
@@ -865,13 +923,6 @@ int journal_load_changesets(journal_t *j, list_t *dst, uint32_t from)
 	local_txn_t(txn, j);
 	txn_begin(txn, false);
 
-	changeset_t *mch = NULL;
-	load_merged_changeset(j, txn, &mch, &from);
-	if (mch != NULL) {
-		add_tail(dst, &mch->n);
-		from = knot_soa_serial(&mch->soa_to->rrs);
-	}
-
 	uint32_t ls = txn->shadow_md.last_serial;
 	iterate(j, txn, load_list_itercb, JOURNAL_ITERATION_CHANGESETS, &dst, from,
 	        ls, normal_iterkeycb);
@@ -880,6 +931,35 @@ int journal_load_changesets(journal_t *j, list_t *dst, uint32_t from)
 	return txn->ret;
 }
 
+int journal_load_chgset_ctx(journal_t *j, chgset_ctx_list_t *dst, uint32_t from)
+{
+	if (j == NULL || j->db == NULL || dst == NULL) return KNOT_EINVAL;
+
+	txn_t *txn = calloc(1, sizeof(*txn) + sizeof(*txn->txn));
+	if (txn == NULL) {
+		return KNOT_ENOMEM;
+	}
+	txn_init(txn, ((void *)txn) + sizeof(*txn), j);
+	txn_begin(txn, false);
+
+	init_list(&dst->l);
+	dst->txn = txn;
+	list_t *dstl = &dst->l;
+
+	uint32_t ls = txn->shadow_md.last_serial;
+	iterate(j, txn, load_list_ctx_itercb, JOURNAL_ITERATION_CHANGESETS, &dstl, from,
+		ls, normal_iterkeycb);
+
+	if (txn->ret != KNOT_EOK) {
+		int ret = txn->ret;
+		txn_commit(txn);
+		free(txn);
+		return ret;
+	}
+
+	return txn->ret;
+}
+
 int load_bootstrap_iterkeycb(iteration_ctx_t *ctx)
 {
 	txn_key_str_u32(ctx->txn, ctx->txn->j->zone, KEY_BOOTSTRAP_CHANGESET, ctx->chunk_index);
@@ -923,7 +1003,7 @@ int journal_load_bootstrap(journal_t *j, list_t *dst)
 		goto jlb_end;
 	}
 	add_tail(dst, &bch->n);
-	uint32_t from = knot_soa_serial(&bch->soa_to->rrs);
+	uint32_t from = knot_soa_serial(bch->soa_to->rrs.rdata);
 
 	uint32_t ls = txn->shadow_md.last_serial;
 	iterate(j, txn, load_list_itercb, JOURNAL_ITERATION_CHANGESETS, &dst,
@@ -991,8 +1071,7 @@ static int delete_merged_changeset(journal_t *j, txn_t *t)
 	txn_check_ret(txn);
 	if (!md_flag(txn, MERGED_SERIAL_VALID)) {
 		txn->ret = KNOT_ENOENT;
-	}
-	else {
+	} else {
 		delete_upto(j, txn, txn->shadow_md.merged_serial, txn->shadow_md.merged_serial);
 	}
 	unreuse_txn(txn, t);
@@ -1207,7 +1286,7 @@ static int merge_unflushed_changesets(journal_t *j, txn_t *_txn, changeset_t **m
 		txn->ret = load_one(j, txn, from, mch);
 		if (!was_merged && was_flushed && txn->ret == KNOT_EOK) {
 			// we have to jump to ONE AFTER last_flushed
-			from = knot_soa_serial(&(*mch)->soa_to->rrs);
+			from = knot_soa_serial((*mch)->soa_to->rrs.rdata);
 			changeset_free(*mch);
 			*mch = NULL;
 			txn->ret = load_one(j, txn, from, mch);
@@ -1216,7 +1295,7 @@ static int merge_unflushed_changesets(journal_t *j, txn_t *_txn, changeset_t **m
 	if (txn->ret != KNOT_EOK) {
 		goto m_u_ch_end;
 	}
-	from = knot_soa_serial(&(*mch)->soa_to->rrs);
+	from = knot_soa_serial((*mch)->soa_to->rrs.rdata);
 
 	if (!serial_equal(from, txn->shadow_md.last_serial_to)) {
 		txn->ret = iterate(j, txn, merge_itercb, JOURNAL_ITERATION_CHANGESETS,
@@ -1232,6 +1311,9 @@ m_u_ch_end:
 	return txn->ret;
 }
 
+dynarray_declare(chunk, knot_db_val_t, DYNARRAY_VISIBILITY_STATIC, 32)
+dynarray_define(chunk, knot_db_val_t, DYNARRAY_VISIBILITY_STATIC)
+
 // uses local context, e.g.: j, txn, changesets, nchs, serialized_size_total, store_changeset_cleanup, inserting_merged
 #define try_flush \
 	if (!md_flushed(txn)) { \
@@ -1261,10 +1343,7 @@ static int store_changesets(journal_t *j, list_t *changesets)
 	size_t nchs = 0, inserted_size = 0, insert_txn_count = 1;
 	size_t serialized_size_changes = 0, serialized_size_merged = 0;
 
-	uint8_t *allchunks = NULL;
-	uint8_t **chunkptrs = NULL;
-	size_t *chunksizes = NULL;
-	knot_db_val_t *vals = NULL;
+	size_t chunks = 0;
 
 	bool inserting_merged = false;
 	bool merged_into_bootstrap = false;
@@ -1329,7 +1408,9 @@ static int store_changesets(journal_t *j, list_t *changesets)
 			delete_tofree(j, txn, tofree, &freed);
 			if (freed < free_min) {
 				txn->ret = KNOT_ESPACE;
-				log_zone_warning(j->zone, "journal, unable to make free space for insert");
+				log_zone_warning(j->zone, "journal, unable to make free space for insert, "
+				                 "required: %"PRIu64", max: %"PRIu64,
+				                 occupied, occupied_max);
 				goto store_changeset_cleanup;
 			}
 		}
@@ -1356,7 +1437,7 @@ static int store_changesets(journal_t *j, list_t *changesets)
 	// PART 4: continuity and duplicity check
 	changeset_t * chs_head = (HEAD(*changesets));
 	bool is_first_bootstrap = (chs_head->soa_from == NULL);
-	uint32_t serial = is_first_bootstrap ? 0 : knot_soa_serial(&chs_head->soa_from->rrs);
+	uint32_t serial = is_first_bootstrap ? 0 : knot_soa_serial(chs_head->soa_from->rrs.rdata);
 	if (md_flag(txn, SERIAL_TO_VALID) && (is_first_bootstrap ||
 	    !serial_equal(txn->shadow_md.last_serial_to, serial)) &&
 	    !inserting_bootstrap /* if inserting bootstrap, drop_journal() was called, so no discontinuity */) {
@@ -1374,7 +1455,7 @@ static int store_changesets(journal_t *j, list_t *changesets)
 		txn_restart(txn);
 	}
 	WALK_LIST(ch, *changesets) {
-		uint32_t serial_to = knot_soa_serial(&ch->soa_to->rrs);
+		uint32_t serial_to = knot_soa_serial(ch->soa_to->rrs.rdata);
 		bool is_this_bootstrap = (ch->soa_from == NULL);
 		bool is_this_merged = (inserting_merged && ch == TAIL(*changesets));
 		if (is_this_bootstrap || is_this_merged) {
@@ -1399,54 +1480,34 @@ static int store_changesets(journal_t *j, list_t *changesets)
 		}
 	}
 
-	// PART 5: serializing into chunks
+	// PART 5: serializing into lmdb
 	WALK_LIST(ch, *changesets) {
 		if (txn->ret != KNOT_EOK) {
 			break;
 		}
 
-		// Twice chsize seems like enough room to store all chunks together.
-		size_t maxchunks = changeset_serialized_size(ch) * 2 / CHUNK_MAX + 1, chunks;
-		allchunks = malloc(maxchunks * CHUNK_MAX);
-		chunkptrs = malloc(maxchunks * sizeof(uint8_t *));
-		chunksizes = malloc(maxchunks * sizeof(size_t));
-		vals = malloc(maxchunks * sizeof(knot_db_val_t));
-		if (allchunks == NULL || chunkptrs == NULL || chunksizes == NULL || vals == NULL) {
+		chunk_dynarray_t dchunks = { 0 };
+		chunks = 0;
+
+		serialize_ctx_t *sctx = serialize_init(ch);
+		if (sctx == NULL) {
 			txn->ret = KNOT_ENOMEM;
 			break;
 		}
-		for (size_t i = 0; i < maxchunks; i++) {
-			chunkptrs[i] = allchunks + i*CHUNK_MAX + JOURNAL_HEADER_SIZE;
-		}
-		txn->ret = changeset_serialize(ch, chunkptrs, CHUNK_MAX - JOURNAL_HEADER_SIZE,
-		                               maxchunks, chunksizes, &chunks);
-		if (txn->ret != KNOT_EOK) {
-			break;
-		}
 
 		bool is_this_merged = (inserting_merged && ch == TAIL(*changesets));
 		bool is_this_bootstrap = (ch->soa_from == NULL);
-		uint32_t serial = is_this_bootstrap ? 0 : knot_soa_serial(&ch->soa_from->rrs);
-		uint32_t serial_to = knot_soa_serial(&ch->soa_to->rrs);
+		uint32_t serial = is_this_bootstrap ? 0 : knot_soa_serial(ch->soa_from->rrs.rdata);
+		uint32_t serial_to = knot_soa_serial(ch->soa_to->rrs.rdata);
 
-		for (size_t i = 0; i < chunks; i++) {
-			vals[i].data = allchunks + i*CHUNK_MAX;
-			vals[i].len = JOURNAL_HEADER_SIZE + chunksizes[i];
-			make_header(vals + i, serial_to, chunks);
-		}
+		while (serialize_unfinished(sctx)) {
+			size_t chunk_size;
+			serialize_prepare(sctx, CHUNK_MAX - JOURNAL_HEADER_SIZE, &chunk_size);
+			if (chunk_size == 0) {
+				break;
+			}
 
-	// PART 6: inserting vals into db
-		for (size_t i = 0; i < chunks; i++) {
-			if (txn->ret != KNOT_EOK) break;
-			if (is_this_bootstrap) {
-				txn_key_str_u32(txn, j->zone, KEY_BOOTSTRAP_CHANGESET, i);
-			}
-			else {
-				txn_key_2u32(txn, j->zone, serial, i);
-			}
-			txn->val = vals[i];
-			txn_insert(txn);
-			inserted_size += (vals+i)->len;
+			inserted_size += chunk_size;
 			if ((float)inserted_size > journal_max_txn(j) * (float)j->db->fslimit) { // insert txn too large
 				inserted_size = 0;
 				txn->shadow_md.dirty_serial = serial;
@@ -1455,8 +1516,33 @@ static int store_changesets(journal_t *j, list_t *changesets)
 				insert_txn_count++;
 				txn->shadow_md.flags &= ~DIRTY_SERIAL_VALID;
 			}
+
+			if (is_this_bootstrap) {
+				txn_key_str_u32(txn, j->zone, KEY_BOOTSTRAP_CHANGESET, chunks);
+			} else {
+				txn_key_2u32(txn, j->zone, serial, chunks);
+			}
+
+			txn->val.data = NULL;
+			txn->val.len = chunk_size + JOURNAL_HEADER_SIZE;
+
+			txn_insert(txn);
+			if (txn->ret != KNOT_EOK) break;
+
+			chunk_dynarray_add(&dchunks, &txn->val);
+
+			chunks++;
+
+			serialize_chunk(sctx, txn->val.data + JOURNAL_HEADER_SIZE, chunk_size);
 		}
 
+		serialize_deinit(sctx);
+
+		dynarray_foreach(chunk, knot_db_val_t, val, dchunks) {
+			make_header(val, serial_to, chunks);
+		}
+		chunk_dynarray_free(&dchunks);
+
 	// PART 7: metadata update
 		if (txn->ret != KNOT_EOK) {
 			break;
@@ -1471,8 +1557,7 @@ static int store_changesets(journal_t *j, list_t *changesets)
 				txn->shadow_md.last_serial_to = serial_to;
 			}
 			txn->shadow_md.flags |= SERIAL_TO_VALID;
-		}
-		else {
+		} else {
 			if (!md_flag(txn, SERIAL_TO_VALID) || md_flag(txn, FIRST_SERIAL_INVALID)) {
 				txn->shadow_md.first_serial = serial;
 			}
@@ -1482,15 +1567,6 @@ static int store_changesets(journal_t *j, list_t *changesets)
 			txn->shadow_md.last_serial_to = serial_to;
 			txn->shadow_md.changeset_count++;
 		}
-
-		free(allchunks);
-		free(chunkptrs);
-		free(chunksizes);
-		free(vals);
-		allchunks = NULL;
-		chunkptrs = NULL;
-		chunksizes = NULL;
-		vals = NULL;
 	}
 
 	// PART X : finalization and cleanup
@@ -1508,11 +1584,6 @@ store_changeset_cleanup:
 		txn_commit(ddtxn);
 	}
 
-	if (allchunks != NULL) free(allchunks);
-	if (chunkptrs != NULL) free(chunkptrs);
-	if (chunksizes != NULL) free(chunksizes);
-	if (vals != NULL) free(vals);
-
 	changeset_t *dbgchst = TAIL(*changesets);
 
 	if (inserting_merged) {
@@ -1762,7 +1833,8 @@ int journal_scrape(journal_t *j)
 	txn_iter_begin(txn);
 	while (txn->ret == KNOT_EOK && txn->iter != NULL) {
 		txn_iter_key(txn, &key);
-		if (knot_dname_is_equal((const knot_dname_t *) key.data, j->zone)) {
+		if (knot_dname_is_equal((const knot_dname_t *) key.data, j->zone)
+		    && key_is_ok(&key, true)) {
 			knot_db_val_t * inskey = dbval_copy(&key);
 			if (inskey == NULL) {
 				txn->ret = KNOT_ENOMEM;
@@ -1798,7 +1870,7 @@ scrape_end:
 
 void journal_metadata_info(journal_t *j, bool *has_bootstrap, kserial_t *merged_serial,
 			   kserial_t *first_serial, kserial_t *last_flushed, kserial_t *serial_to,
-			   uint64_t *occupied)
+			   uint64_t *occupied, uint64_t *occupied_all_zones)
 {
 	// NOTE: there is NEVER the situation that only merged changeset would be present and no common changeset in db.
 
@@ -1824,6 +1896,8 @@ void journal_metadata_info(journal_t *j, bool *has_bootstrap, kserial_t *merged_
 		return;
 	}
 
+	uint64_t occupied_total = knot_db_lmdb_get_usage(j->db->db);
+
 	local_txn_t(txn, j);
 	txn_begin(txn, false);
 	txn_check_open(txn);
@@ -1854,10 +1928,13 @@ void journal_metadata_info(journal_t *j, bool *has_bootstrap, kserial_t *merged_
 		if (last_inserter != NULL && knot_dname_is_equal(last_inserter, j->zone)) {
 			size_t lz_occupied;
 			md_get_common_last_occupied(txn, &lz_occupied);
-			*occupied += knot_db_lmdb_get_usage(j->db->db) - lz_occupied;
+			*occupied += occupied_total - lz_occupied;
 		}
 		free(last_inserter);
 	}
+	if (occupied_all_zones != NULL) {
+		*occupied_all_zones = occupied_total;
+	}
 
 	txn_abort(txn);
 }
@@ -1894,7 +1971,11 @@ int journal_db_list_zones(journal_db_t **db, list_t *zones)
 		if (txn->ret == KNOT_EOK && *compare_metaflag == '\0') {
 			compare_metaflag -= metaflag_len;
 			if (strcmp(compare_metaflag, MDKEY_PERZONE_FLAGS) == 0) {
-				char *found_zone = knot_dname_to_str_alloc((const knot_dname_t *)key.data);
+				knot_dname_t *found_zone = knot_dname_copy((const knot_dname_t *)key.data, NULL);
+				if (found_zone == NULL) {
+					txn->ret = KNOT_ENOMEM;
+					break;
+				}
 				ptrlist_add(zones, found_zone, NULL);
 			}
 		}
@@ -2004,17 +2085,15 @@ int journal_check(journal_t *j, journal_check_level_t warn_level)
 		ret = load_bootstrap_changeset(j, txn, &ch);
 		if (ret != KNOT_EOK) {
 			jch_warn("can't read bootstrap changeset (%s)", knot_strerror(ret));
-		}
-		else {
+		} else {
 			changeset_free(ch);
 		}
 		goto check_merged;
-	}
-	else {
+	} else {
 		ret = load_bootstrap_changeset(j, txn, &ch);
 		switch (ret) {
 		case KNOT_EOK:
-			sto = knot_soa_serial(&ch->soa_to->rrs);
+			sto = knot_soa_serial(ch->soa_to->rrs.rdata);
 			jch_info("bootstrap changeset loaded, sto %u", sto);
 			changeset_free(ch);
 			break;
@@ -2034,7 +2113,7 @@ int journal_check(journal_t *j, journal_check_level_t warn_level)
 		goto check_merged;
 	}
 
-	sfrom = knot_soa_serial(&ch->soa_from->rrs), sto = knot_soa_serial(&ch->soa_to->rrs);
+	sfrom = knot_soa_serial(ch->soa_from->rrs.rdata), sto = knot_soa_serial(ch->soa_to->rrs.rdata);
 	if (!serial_equal(txn->shadow_md.first_serial, sfrom)) {
 		jch_warn("first changeset's serial 'from' %u is not ok", sfrom);
 	}
@@ -2045,9 +2124,8 @@ int journal_check(journal_t *j, journal_check_level_t warn_level)
 		if (ret != KNOT_EOK) {
 			jch_warn("can't read last flushed changeset %u (%s)",
 			         txn->shadow_md.last_flushed, knot_strerror(ret));
-		}
-		else {
-			first_unflushed = knot_soa_serial(&ch->soa_to->rrs);
+		} else {
+			first_unflushed = knot_soa_serial(ch->soa_to->rrs.rdata);
 		}
 	}
 	if (ret == KNOT_EOK) {
@@ -2061,9 +2139,8 @@ int journal_check(journal_t *j, journal_check_level_t warn_level)
 	ret = load_one(j, txn, sto, &ch);
 	if (ret != KNOT_EOK) {
 		jch_warn("can't read second changeset %u (%s)", sto, knot_strerror(ret));
-	}
-	else {
-		sfrom = knot_soa_serial(&ch->soa_from->rrs);
+	} else {
+		sfrom = knot_soa_serial(ch->soa_from->rrs.rdata);
 		if (!serial_equal(sfrom, sto)) {
 			jch_warn("second changeset's serial 'from' %u is not ok", sfrom);
 		}
@@ -2088,14 +2165,14 @@ int journal_check(journal_t *j, journal_check_level_t warn_level)
 	}
 
 	ch = HEAD(l);
-	if (!serial_equal(sfrom, knot_soa_serial(&ch->soa_from->rrs))) {
+	if (!serial_equal(sfrom, knot_soa_serial(ch->soa_from->rrs.rdata))) {
 		jch_warn("first listed changeset's serial 'from' %u is not ok",
-		         knot_soa_serial(&ch->soa_from->rrs));
+		         knot_soa_serial(ch->soa_from->rrs.rdata));
 	}
 	ch = TAIL(l);
-	if (!serial_equal(sto, knot_soa_serial(&ch->soa_to->rrs))) {
+	if (!serial_equal(sto, knot_soa_serial(ch->soa_to->rrs.rdata))) {
 		jch_warn("last listed changeset's serial 'to' %u is not ok",
-		         knot_soa_serial(&ch->soa_to->rrs));
+		         knot_soa_serial(ch->soa_to->rrs.rdata));
 	}
 	changesets_free(&l);
 
@@ -2108,10 +2185,9 @@ check_merged:
 		ret = load_merged_changeset(j, txn, &ch, NULL);
 		if (ret != KNOT_EOK) {
 			jch_warn("can't read merged changeset (%s)", knot_strerror(ret));
-		}
-		else {
-			sfrom = knot_soa_serial(&ch->soa_from->rrs);
-			sto = knot_soa_serial(&ch->soa_to->rrs);
+		} else {
+			sfrom = knot_soa_serial(ch->soa_from->rrs.rdata);
+			sto = knot_soa_serial(ch->soa_to->rrs.rdata);
 			jch_info("merged changeset %u -> %u (size %zu)", sfrom, sto,
 			         changeset_serialized_size(ch));
 			if (!serial_equal(sfrom, txn->shadow_md.merged_serial)) {
