@@ -20,6 +20,7 @@
 #include <stdlib.h>
 #include <string.h>
 
+#include "contrib/sockaddr.h"
 #include "libknot/attribute.h"
 #include "knot/common/log.h"
 #include "knot/conf/module.h"
@@ -498,6 +499,34 @@ knotd_conf_t knotd_conf_check_item(knotd_conf_check_args_t *args,
 	return out;
 }
 
+_public_
+bool knotd_conf_addr_range_match(const knotd_conf_t *range,
+                                 const struct sockaddr_storage *addr)
+{
+	if (range == NULL || addr == NULL) {
+		return false;
+	}
+
+	for (size_t i = 0; i < range->count; i++) {
+		knotd_conf_val_t *val = &range->multi[i];
+		if (val->addr_max.ss_family == AF_UNSPEC) {
+			if (sockaddr_net_match((struct sockaddr *)addr,
+			                       (struct sockaddr *)&val->addr,
+			                       val->addr_mask)) {
+				return true;
+			}
+		} else {
+			if (sockaddr_range_match((struct sockaddr *)addr,
+			                         (struct sockaddr *)&val->addr,
+			                         (struct sockaddr *)&val->addr_max)) {
+				return true;
+			}
+		}
+	}
+
+	return false;
+}
+
 _public_
 void knotd_conf_free(knotd_conf_t *conf)
 {
@@ -594,7 +623,7 @@ int knotd_mod_dnssec_sign_rrset(knotd_mod_t *mod, knot_rrset_t *rrsigs,
 		}
 
 		int ret = knot_sign_rrset(rrsigs, rrset, key->key, key->ctx,
-		                          mod->dnssec, mm);
+		                          mod->dnssec, mm, NULL);
 		if (ret != KNOT_EOK) {
 			return ret;
 		}
