@@ -1,4 +1,4 @@
-/*  Copyright (C) 2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -42,6 +42,7 @@ static int rr_data_from(const knot_rrset_t *rrset, struct rr_data *data, knot_mm
 	if (ret != KNOT_EOK) {
 		return ret;
 	}
+	data->ttl = rrset->ttl;
 	data->type = rrset->type;
 	data->additional = NULL;
 
@@ -73,16 +74,13 @@ static int add_rrset_no_merge(zone_node_t *node, const knot_rrset_t *rrset,
 }
 
 /*! \brief Checks if the added RR has the same TTL as the first RR in the node. */
-static bool ttl_error(struct rr_data *node_data, const knot_rrset_t *rrset)
+static bool ttl_changed(struct rr_data *node_data, const knot_rrset_t *rrset)
 {
-	if (rrset->type == KNOT_RRTYPE_RRSIG || node_data->rrs.rr_count == 0) {
+	if (rrset->type == KNOT_RRTYPE_RRSIG || node_data->rrs.count == 0) {
 		return false;
 	}
 
-	const uint32_t inserted_ttl = knot_rdataset_ttl(&rrset->rrs);
-	const uint32_t node_ttl = knot_rdataset_ttl(&node_data->rrs);
-	// Return error if TTLs don't match.
-	return inserted_ttl != node_ttl;
+	return rrset->ttl != node_data->ttl;
 }
 
 zone_node_t *node_new(const knot_dname_t *owner, knot_mm_t *mm)
@@ -122,20 +120,19 @@ void node_free_rrsets(zone_node_t *node, knot_mm_t *mm)
 	node->rrset_count = 0;
 }
 
-void node_free(zone_node_t **node, knot_mm_t *mm)
+void node_free(zone_node_t *node, knot_mm_t *mm)
 {
-	if (node == NULL || *node == NULL) {
+	if (node == NULL) {
 		return;
 	}
 
-	if ((*node)->rrs != NULL) {
-		mm_free(mm, (*node)->rrs);
+	knot_dname_free(node->owner, mm);
+
+	if (node->rrs != NULL) {
+		mm_free(mm, node->rrs);
 	}
 
-	knot_dname_free(&(*node)->owner, mm);
-
-	mm_free(mm, *node);
-	*node = NULL;
+	mm_free(mm, node);
 }
 
 zone_node_t *node_shallow_copy(const zone_node_t *src, knot_mm_t *mm)
@@ -157,7 +154,7 @@ zone_node_t *node_shallow_copy(const zone_node_t *src, knot_mm_t *mm)
 	size_t rrlen = sizeof(struct rr_data) * src->rrset_count;
 	dst->rrs = mm_alloc(mm, rrlen);
 	if (dst->rrs == NULL) {
-		node_free(&dst, mm);
+		node_free(dst, mm);
 		return NULL;
 	}
 	memcpy(dst->rrs, src->rrs, rrlen);
@@ -179,10 +176,9 @@ int node_add_rrset(zone_node_t *node, const knot_rrset_t *rrset, knot_mm_t *mm)
 	for (uint16_t i = 0; i < node->rrset_count; ++i) {
 		if (node->rrs[i].type == rrset->type) {
 			struct rr_data *node_data = &node->rrs[i];
-			const bool ttl_err = ttl_error(node_data, rrset);
-			if (ttl_err) {
-				knot_rdataset_set_ttl(&node_data->rrs,
-				                      knot_rdataset_ttl(&rrset->rrs));
+			const bool ttl_change = ttl_changed(node_data, rrset);
+			if (ttl_change) {
+				node_data->ttl = rrset->ttl;
 			}
 
 			int ret = knot_rdataset_merge(&node_data->rrs,
@@ -190,7 +186,7 @@ int node_add_rrset(zone_node_t *node, const knot_rrset_t *rrset, knot_mm_t *mm)
 			if (ret != KNOT_EOK) {
 				return ret;
 			} else {
-				return ttl_err ? KNOT_ETTL : KNOT_EOK;
+				return ttl_change ? KNOT_ETTL : KNOT_EOK;
 			}
 		}
 	}
@@ -207,6 +203,8 @@ void node_remove_rdataset(zone_node_t *node, uint16_t type)
 
 	for (int i = 0; i < node->rrset_count; ++i) {
 		if (node->rrs[i].type == type) {
+			// We need to free additionals from this rr_data before it gets overwritten.
+			additional_clear(node->rrs[i].additional);
 			memmove(node->rrs + i, node->rrs + i + 1,
 			        (node->rrset_count - i - 1) * sizeof(struct rr_data));
 			--node->rrset_count;
@@ -276,28 +274,18 @@ bool node_rrtype_is_signed(const zone_node_t *node, uint16_t type)
 		return false;
 	}
 
-	uint16_t rrsigs_rdata_count = rrsigs->rr_count;
+	uint16_t rrsigs_rdata_count = rrsigs->count;
+	knot_rdata_t *rrsig = rrsigs->rdata;
 	for (uint16_t i = 0; i < rrsigs_rdata_count; ++i) {
-		const uint16_t type_covered =
-			knot_rrsig_type_covered(rrsigs, i);
-		if (type_covered == type) {
+		if (knot_rrsig_type_covered(rrsig) == type) {
 			return true;
 		}
+		rrsig = knot_rdataset_next(rrsig);
 	}
 
 	return false;
 }
 
-bool node_rrtype_exists(const zone_node_t *node, uint16_t type)
-{
-	return node_rdataset(node, type) != NULL;
-}
-
-bool node_empty(const zone_node_t *node)
-{
-	return node == NULL || node->rrset_count == 0;
-}
-
 bool node_bitmap_equal(const zone_node_t *a, const zone_node_t *b)
 {
 	if (a == NULL || b == NULL || a->rrset_count != b->rrset_count) {
