@@ -1,4 +1,4 @@
-/*  Copyright (C) 2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -40,6 +40,7 @@ typedef struct zone_node {
 	 */
 	struct zone_node *prev;
 	struct zone_node *nsec3_node; /*! NSEC3 node corresponding to this node. */
+	struct zone_node *nsec3_wildcard_prev; /*! NSEC3 node for proof of wildcard non-existence. */
 	uint32_t children; /*!< Count of children nodes in DNS hierarchy. */
 	uint16_t rrset_count; /*!< Number of RRSets stored in the node. */
 	uint8_t flags; /*!< \ref node_flags enum. */
@@ -60,6 +61,7 @@ typedef struct {
 
 /*!< \brief Structure storing RR data. */
 struct rr_data {
+	uint32_t ttl; /*!< RRSet TTL. */
 	uint16_t type; /*!< RR type of data. */
 	knot_rdataset_t rrs; /*!< Data of given type. */
 	additional_t *additional; /*!< Additional nodes with glues. */
@@ -78,7 +80,9 @@ enum node_flags {
 	/*! \brief Node is empty and will be deleted after update. */
 	NODE_FLAGS_EMPTY =           1 << 3,
 	/*! \brief Node has a wildcard child. */
-	NODE_FLAGS_WILDCARD_CHILD =  1 << 4
+	NODE_FLAGS_WILDCARD_CHILD =  1 << 4,
+	/*! \brief Is this NSEC3 node compatible with zone's NSEC3PARAMS ? */
+	NODE_FLAGS_IN_NSEC3_CHAIN =  1 << 5,
 };
 
 /*!
@@ -111,12 +115,11 @@ void node_free_rrsets(zone_node_t *node, knot_mm_t *mm);
  * \brief Destroys the node structure.
  *
  * Does not destroy the data within the node.
- * Also sets the given pointer to NULL.
  *
  * \param node  Node to be destroyed.
  * \param mm    Memory context to use.
  */
-void node_free(zone_node_t **node, knot_mm_t *mm);
+void node_free(zone_node_t *node, knot_mm_t *mm);
 
 /*!
  * \brief Creates a shallow copy of node structure, RR data are shared.
@@ -137,6 +140,7 @@ zone_node_t *node_shallow_copy(const zone_node_t *src, knot_mm_t *mm);
  * \param mm       Memory context to use.
  *
  * \return KNOT_E*
+ * \retval KNOT_ETTL  RRSet TTL was updated.
  */
 int node_add_rrset(zone_node_t *node, const knot_rrset_t *rrset, knot_mm_t *mm);
 
@@ -195,7 +199,10 @@ bool node_rrtype_is_signed(const zone_node_t *node, uint16_t type);
  *
  * \return True/False.
  */
-bool node_rrtype_exists(const zone_node_t *node, uint16_t type);
+inline static bool node_rrtype_exists(const zone_node_t *node, uint16_t type)
+{
+	return node_rdataset(node, type) != NULL;
+}
 
 /*!
  * \brief Checks whether node is empty. Node is empty when NULL or when no
@@ -205,7 +212,10 @@ bool node_rrtype_exists(const zone_node_t *node, uint16_t type);
  *
  * \return True/False.
  */
-bool node_empty(const zone_node_t *node);
+inline static bool node_empty(const zone_node_t *node)
+{
+	return node == NULL || node->rrset_count == 0;
+}
 
 /*!
  * \brief Check whether two nodes have equal set of rrtypes.
@@ -231,7 +241,8 @@ static inline knot_rrset_t node_rrset(const zone_node_t *node, uint16_t type)
 	for (uint16_t i = 0; node && i < node->rrset_count; ++i) {
 		if (node->rrs[i].type == type) {
 			struct rr_data *rr_data = &node->rrs[i];
-			knot_rrset_init(&rrset, node->owner, type, KNOT_CLASS_IN);
+			knot_rrset_init(&rrset, node->owner, type, KNOT_CLASS_IN,
+			                rr_data->ttl);
 			rrset.rrs = rr_data->rrs;
 			rrset.additional = rr_data->additional;
 			return rrset;
@@ -259,7 +270,8 @@ static inline knot_rrset_t node_rrset_at(const zone_node_t *node, size_t pos)
 	}
 
 	struct rr_data *rr_data = &node->rrs[pos];
-	knot_rrset_init(&rrset, node->owner, rr_data->type, KNOT_CLASS_IN);
+	knot_rrset_init(&rrset, node->owner, rr_data->type, KNOT_CLASS_IN,
+	                rr_data->ttl);
 	rrset.rrs = rr_data->rrs;
 	rrset.additional = rr_data->additional;
 	return rrset;
