@@ -1,4 +1,4 @@
-/*  Copyright (C) 2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -13,21 +13,9 @@
     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
-/*!
- * \file
- *
- * \brief Zone structure and API for manipulating it.
- *
- * \addtogroup zone
- * @{
- */
 
 #pragma once
 
-#include <time.h>
-#include <stdbool.h>
-#include <stdint.h>
-
 #include "knot/conf/conf.h"
 #include "knot/conf/confio.h"
 #include "knot/journal/journal.h"
@@ -37,7 +25,6 @@
 #include "libknot/dname.h"
 #include "libknot/packet/pkt.h"
 
-struct process_query_param;
 struct zone_update;
 
 /*!
@@ -66,6 +53,7 @@ typedef struct zone
 		time_t mtime;
 		uint32_t serial;
 		bool exists;
+		bool resigned;
 	} zonefile;
 
 	/*! \brief Zone events. */
@@ -125,8 +113,8 @@ void zone_free(zone_t **zone_ptr);
 void zone_control_clear(zone_t *zone);
 
 int zone_change_store(conf_t *conf, zone_t *zone, changeset_t *change);
-int zone_changes_store(conf_t *conf, zone_t *zone, list_t *chgs);
 int zone_changes_load(conf_t *conf, zone_t *zone, list_t *dst, uint32_t from);
+int zone_chgset_ctx_load(conf_t *conf, zone_t *zone, chgset_ctx_list_t *dst, uint32_t from);
 int zone_in_journal_load(conf_t *conf, zone_t *zone, list_t *dst);
 int zone_in_journal_store(conf_t *conf, zone_t *zone, zone_contents_t *new_contents);
 int zone_journal_serial(conf_t *conf, zone_t *zone, bool *is_empty, uint32_t *serial_to);
@@ -186,5 +174,3 @@ int zone_get_master_serial(zone_t *zone, uint32_t *serial);
 int zone_set_lastsigned_serial(zone_t *zone, uint32_t serial);
 
 bool zone_get_lastsigned_serial(zone_t *zone, uint32_t *serial);
-
-/*! @} */
