@@ -1,4 +1,4 @@
-/*  Copyright (C) 2016 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,13 +18,13 @@
 #include <stdlib.h>
 #include <time.h>
 
-#include "dnssec/random.h"
+#include "libdnssec/random.h"
 #include "utils/common/exec.h"
 #include "utils/common/msg.h"
 #include "utils/common/netio.h"
 #include "utils/common/params.h"
 #include "libknot/libknot.h"
-#include "contrib/print.h"
+#include "contrib/ctype.h"
 #include "contrib/sockaddr.h"
 #include "contrib/openbsd/strlcat.h"
 #include "contrib/ucw/lists.h"
@@ -186,12 +186,35 @@ static void print_footer(const size_t total_len,
 	}
 }
 
-static void print_edns_client_subnet(const uint8_t *data, const uint16_t len)
+static void print_hex(const uint8_t *data, uint16_t len)
+{
+	for (int i = 0; i < len; i++) {
+		printf("%02X", data[i]);
+	}
+}
+
+static void print_nsid(const uint8_t *data, uint16_t len)
+{
+	if (len == 0) {
+		return;
+	}
+
+	print_hex(data, len);
+
+	// Check if printable string.
+	for (int i = 0; i < len; i++) {
+		if (!is_print(data[i])) {
+			return;
+		}
+	}
+	printf(" \"%.*s\"", len, data);
+}
+
+static void print_edns_client_subnet(const uint8_t *data, uint16_t len)
 {
 	knot_edns_client_subnet_t ecs = { 0 };
 	int ret = knot_edns_client_subnet_parse(&ecs, data, len);
 	if (ret != KNOT_EOK) {
-		printf("\n");
 		return;
 	}
 
@@ -202,7 +225,7 @@ static void print_edns_client_subnet(const uint8_t *data, const uint16_t len)
 	char addr_str[SOCKADDR_STRLEN] = { 0 };
 	sockaddr_tostr(addr_str, sizeof(addr_str), (struct sockaddr *)&addr);
 
-	printf("%s/%u/%u\n", addr_str, ecs.source_len, ecs.scope_len);
+	printf("%s/%u/%u", addr_str, ecs.source_len, ecs.scope_len);
 }
 
 static void print_section_opt(const knot_pkt_t *packet)
@@ -230,9 +253,9 @@ static void print_section_opt(const knot_pkt_t *packet)
 	       knot_edns_get_payload(packet->opt_rr),
 	       ercode_str);
 
-	knot_rdata_t *rdata = knot_rdataset_at(&packet->opt_rr->rrs, 0);
-	wire_ctx_t wire = wire_ctx_init_const(knot_rdata_data(rdata),
-	                                      knot_rdata_rdlen(rdata));
+	assert(packet->opt_rr->rrs.count > 0);
+	knot_rdata_t *rdata = packet->opt_rr->rrs.rdata;
+	wire_ctx_t wire = wire_ctx_init_const(rdata->data, rdata->len);
 
 	while (wire_ctx_available(&wire) >= KNOT_EDNS_OPTION_HDRLEN) {
 		uint16_t opt_code = wire_ctx_read_u16(&wire);
@@ -247,23 +270,24 @@ static void print_section_opt(const knot_pkt_t *packet)
 		switch (opt_code) {
 		case KNOT_EDNS_OPTION_NSID:
 			printf(";; NSID: ");
-			short_hex_print(opt_data, opt_len);
-			if (opt_len > 0) {
-				printf(";;     :  ");
-				txt_print(opt_data, opt_len);
-			}
+			print_nsid(opt_data, opt_len);
 			break;
 		case KNOT_EDNS_OPTION_CLIENT_SUBNET:
 			printf(";; CLIENT-SUBNET: ");
 			print_edns_client_subnet(opt_data, opt_len);
 			break;
 		case KNOT_EDNS_OPTION_PADDING:
-			printf(";; PADDING: %u B\n", opt_len);
+			printf(";; PADDING: %u B", opt_len);
+			break;
+		case KNOT_EDNS_OPTION_COOKIE:
+			printf(";; COOKIE: ");
+			print_hex(opt_data, opt_len);
 			break;
 		default:
 			printf(";; Option (%u): ", opt_code);
-			short_hex_print(opt_data, opt_len);
+			print_hex(opt_data, opt_len);
 		}
+		printf("\n");
 
 		wire_ctx_skip(&wire, opt_len);
 	}
@@ -285,7 +309,7 @@ static void print_section_question(const knot_dname_t *owner,
 	knot_dump_style_t qstyle = style->style;
 	qstyle.empty_ttl = true;
 
-	knot_rrset_t *question = knot_rrset_new(owner, qtype, qclass, NULL);
+	knot_rrset_t *question = knot_rrset_new(owner, qtype, qclass, 0, NULL);
 
 	if (knot_rrset_txt_dump_header(question, 0, buf, buflen, &qstyle) < 0) {
 		WARN("can't print whole question section\n");
@@ -293,7 +317,7 @@ static void print_section_question(const knot_dname_t *owner,
 
 	printf("%s\n", buf);
 
-	knot_rrset_free(&question, NULL);
+	knot_rrset_free(question, NULL);
 	free(buf);
 }
 
@@ -336,7 +360,7 @@ static void print_section_dig(const knot_rrset_t *rrsets,
 
 	for (size_t i = 0; i < count; i++) {
 		const knot_rrset_t *rrset = &rrsets[i];
-		uint16_t rrset_rdata_count = rrset->rrs.rr_count;
+		uint16_t rrset_rdata_count = rrset->rrs.count;
 		for (uint16_t j = 0; j < rrset_rdata_count; j++) {
 			while (knot_rrset_txt_dump_data(rrset, j, buf, buflen,
 			                                &(style->style)) < 0) {
@@ -380,7 +404,7 @@ static void print_section_host(const knot_rrset_t *rrsets,
 		}
 		descr = knot_lookup_by_id(rtypes, rrset->type);
 
-		uint16_t rrset_rdata_count = rrset->rrs.rr_count;
+		uint16_t rrset_rdata_count = rrset->rrs.count;
 		for (uint16_t j = 0; j < rrset_rdata_count; j++) {
 			if (rrset->type == KNOT_RRTYPE_CNAME &&
 			    style->hide_cname) {
