@@ -1,4 +1,4 @@
-/*  Copyright (C) 2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -235,7 +235,7 @@ static void process_dnstap(const query_t *query)
 			ERR("can't print dnstap message\n");
 		}
 
-		knot_pkt_free(&pkt);
+		knot_pkt_free(pkt);
 		dt_reader_free_frame(reader, &frame);
 	}
 }
@@ -266,11 +266,11 @@ static int add_query_edns(knot_pkt_t *packet, const query_t *query, uint16_t max
 	}
 
 	/* Append EDNS-client-subnet. */
-	if (query->subnet != NULL) {
-		size_t size = knot_edns_client_subnet_size(query->subnet);
+	if (query->subnet.family != AF_UNSPEC) {
+		uint16_t size = knot_edns_client_subnet_size(&query->subnet);
 		uint8_t data[size];
 
-		ret = knot_edns_client_subnet_write(data, size, query->subnet);
+		ret = knot_edns_client_subnet_write(data, size, &query->subnet);
 		if (ret != KNOT_EOK) {
 			knot_rrset_clear(&opt_rr, &packet->mm);
 			return ret;
@@ -284,6 +284,25 @@ static int add_query_edns(knot_pkt_t *packet, const query_t *query, uint16_t max
 		}
 	}
 
+	/* Append a cookie option if present. */
+	if (query->cc.len > 0) {
+		uint16_t size = knot_edns_cookie_size(&query->cc, &query->sc);
+		uint8_t data[size];
+
+		ret = knot_edns_cookie_write(data, size, &query->cc, &query->sc);
+		if (ret != KNOT_EOK) {
+			knot_rrset_clear(&opt_rr, &packet->mm);
+			return ret;
+		}
+
+		ret = knot_edns_add_option(&opt_rr, KNOT_EDNS_OPTION_COOKIE,
+		                           size, data, &packet->mm);
+		if (ret != KNOT_EOK) {
+			knot_rrset_clear(&opt_rr, &packet->mm);
+			return ret;
+		}
+	}
+
 	/* Append EDNS Padding. */
 	int padding = query->padding;
 	if (padding != -3 && query->alignment > 0) {
@@ -291,7 +310,7 @@ static int add_query_edns(knot_pkt_t *packet, const query_t *query, uint16_t max
 		                                   knot_rrset_size(&opt_rr),
 		                                   query->alignment);
 	} else if (query->padding == -2 || (query->padding == -1 && query->tls.enable)) {
-		padding = knot_edns_default_padding_size(packet, &opt_rr);
+		padding = knot_pkt_default_padding_size(packet, &opt_rr);
 	}
 	if (padding > -1) {
 		uint8_t zeros[padding];
@@ -305,6 +324,18 @@ static int add_query_edns(knot_pkt_t *packet, const query_t *query, uint16_t max
 		}
 	}
 
+	/* Append custom EDNS options. */
+	node_t *node = NULL;
+	WALK_LIST(node, query->edns_opts) {
+		ednsopt_t *opt = (ednsopt_t *)node;
+		ret = knot_edns_add_option(&opt_rr, opt->code, opt->length,
+		                           opt->data, &packet->mm);
+		if (ret != KNOT_EOK) {
+			knot_rrset_clear(&opt_rr, &packet->mm);
+			return ret;
+		}
+	}
+
 	/* Add prepared OPT to packet. */
 	ret = knot_pkt_put(packet, KNOT_COMPR_HINT_NONE, &opt_rr, KNOT_PF_FREE);
 	if (ret != KNOT_EOK) {
@@ -325,8 +356,9 @@ static bool do_padding(const query_t *query)
 static bool use_edns(const query_t *query)
 {
 	return query->edns > -1 || query->udp_size > -1 || query->nsid ||
-	       query->flags.do_flag || query->subnet != NULL ||
-	       do_padding(query);
+	       query->flags.do_flag || query->subnet.family != AF_UNSPEC ||
+	       query->cc.len > 0 || do_padding(query) ||
+	       !ednsopt_list_empty(&query->edns_opts);
 }
 
 static knot_pkt_t *create_query_packet(const query_t *query)
@@ -377,20 +409,16 @@ static knot_pkt_t *create_query_packet(const query_t *query)
 		knot_wire_set_opcode(packet->wire, KNOT_OPCODE_NOTIFY);
 	}
 
-	// Create QNAME from string.
+	// Set packet question if available.
 	knot_dname_t *qname = knot_dname_from_str_alloc(query->owner);
-	if (qname == NULL) {
-		knot_pkt_free(&packet);
-		return NULL;
-	}
-
-	// Set packet question.
-	int ret = knot_pkt_put_question(packet, qname, query->class_num,
-	                                query->type_num);
-	if (ret != KNOT_EOK) {
-		knot_dname_free(&qname, NULL);
-		knot_pkt_free(&packet);
-		return NULL;
+	if (qname != NULL) {
+		int ret = knot_pkt_put_question(packet, qname, query->class_num,
+		                                query->type_num);
+		if (ret != KNOT_EOK) {
+			knot_dname_free(qname, NULL);
+			knot_pkt_free(packet);
+			return NULL;
+		}
 	}
 
 	// For IXFR query or NOTIFY query with SOA serial, add a proper section.
@@ -408,35 +436,35 @@ static knot_pkt_t *create_query_packet(const query_t *query)
 		knot_rrset_t *soa = knot_rrset_new(qname,
 		                                   KNOT_RRTYPE_SOA,
 		                                   query->class_num,
+		                                   0,
 		                                   &packet->mm);
-		knot_dname_free(&qname, NULL);
+		knot_dname_free(qname, NULL);
 		if (soa == NULL) {
-			knot_pkt_free(&packet);
+			knot_pkt_free(packet);
 			return NULL;
 		}
 
 		// Fill in blank SOA rdata to rrset.
-		ret = knot_rrset_add_rdata(soa, wire, sizeof(wire), 0,
-		                           &packet->mm);
+		int ret = knot_rrset_add_rdata(soa, wire, sizeof(wire), &packet->mm);
 		if (ret != KNOT_EOK) {
-			knot_rrset_free(&soa, &packet->mm);
-			knot_pkt_free(&packet);
+			knot_rrset_free(soa, &packet->mm);
+			knot_pkt_free(packet);
 			return NULL;
 		}
 
 		// Set SOA serial.
-		knot_soa_serial_set(&soa->rrs, query->serial);
+		knot_soa_serial_set(soa->rrs.rdata, query->serial);
 
 		ret = knot_pkt_put(packet, KNOT_COMPR_HINT_NONE, soa, KNOT_PF_FREE);
 		if (ret != KNOT_EOK) {
-			knot_rrset_free(&soa, &packet->mm);
-			knot_pkt_free(&packet);
+			knot_rrset_free(soa, &packet->mm);
+			knot_pkt_free(packet);
 			return NULL;
 		}
 
 		free(soa);
 	} else {
-		knot_dname_free(&qname, NULL);
+		knot_dname_free(qname, NULL);
 	}
 
 	// Begin additional section
@@ -447,7 +475,7 @@ static knot_pkt_t *create_query_packet(const query_t *query)
 		int ret = add_query_edns(packet, query, max_size);
 		if (ret != KNOT_EOK) {
 			ERR("can't set up EDNS section\n");
-			knot_pkt_free(&packet);
+			knot_pkt_free(packet);
 			return NULL;
 		}
 	}
@@ -485,12 +513,9 @@ static void check_reply_question(const knot_pkt_t *reply,
 		return;
 	}
 
-	int name_diff = knot_dname_cmp(knot_pkt_qname(reply),
-	                               knot_pkt_qname(query));
-
-	if (knot_pkt_qclass(reply) != knot_pkt_qclass(query) ||
-	    knot_pkt_qtype(reply)  != knot_pkt_qtype(query) ||
-	    name_diff != 0) {
+	if (!knot_dname_is_equal(knot_pkt_qname(reply), knot_pkt_qname(query)) ||
+	    knot_pkt_qclass(reply) != knot_pkt_qclass(query) ||
+	    knot_pkt_qtype(reply)  != knot_pkt_qtype(query)) {
 		WARN("query/response question sections are different\n");
 		return;
 	}
@@ -509,7 +534,7 @@ static int64_t first_serial_check(const knot_pkt_t *reply)
 	if (first->type != KNOT_RRTYPE_SOA) {
 		return -1;
 	} else {
-		return knot_soa_serial(&first->rrs);
+		return knot_soa_serial(first->rrs.rdata);
 	}
 }
 
@@ -529,7 +554,7 @@ static bool finished_xfr(const uint32_t serial, const knot_pkt_t *reply,
 	} else if (answer->count == 1 && msg_count == 1) {
 		return is_ixfr;
 	} else {
-		return knot_soa_serial(&last->rrs) == serial;
+		return knot_soa_serial(last->rrs.rdata) == serial;
 	}
 }
 
@@ -607,7 +632,7 @@ static int process_query_packet(const knot_pkt_t      *query,
 			} else {
 				ERR("can't print query packet\n");
 			}
-			knot_pkt_free(&q);
+			knot_pkt_free(q);
 		} else {
 			ERR("can't print query packet\n");
 		}
@@ -646,7 +671,7 @@ static int process_query_packet(const knot_pkt_t      *query,
 		// Parse reply to the packet structure.
 		if (knot_pkt_parse(reply, KNOT_PF_NOCANON) != KNOT_EOK) {
 			ERR("malformed reply packet from %s\n", net->remote_str);
-			knot_pkt_free(&reply);
+			knot_pkt_free(reply);
 			net_close(net);
 			return 0;
 		}
@@ -656,20 +681,21 @@ static int process_query_packet(const knot_pkt_t      *query,
 			break;
 		// Check for timeout.
 		} else if (time_diff_ms(&t_query, &t_end) > 1000 * net->wait) {
-			knot_pkt_free(&reply);
+			knot_pkt_free(reply);
 			net_close(net);
 			return -1;
 		}
 
-		knot_pkt_free(&reply);
+		knot_pkt_free(reply);
 	}
 
 	// Check for TC bit and repeat query with TCP if required.
 	if (knot_wire_get_tc(reply->wire) != 0 &&
 	    ignore_tc == false && net->socktype == SOCK_DGRAM) {
-		WARN("truncated reply from %s, retrying over TCP\n",
+		printf("\n");
+		WARN("truncated reply from %s, retrying over TCP\n\n",
 		     net->remote_str);
-		knot_pkt_free(&reply);
+		knot_pkt_free(reply);
 		net_close(net);
 
 		net->socktype = SOCK_STREAM;
@@ -697,7 +723,46 @@ static int process_query_packet(const knot_pkt_t      *query,
 		}
 	}
 
-	knot_pkt_free(&reply);
+	// Check for BADCOOKIE RCODE and repeat query with the new cookie if required.
+	if (knot_pkt_ext_rcode(reply) == KNOT_RCODE_BADCOOKIE && query_ctx->badcookie) {
+		printf("\n");
+		WARN("bad cookie from %s, retrying with the received one\n",
+		     net->remote_str);
+		net_close(net);
+
+		// Prepare new query context.
+		query_t new_ctx = *query_ctx;
+
+		uint8_t *opt = knot_pkt_edns_option(reply, KNOT_EDNS_OPTION_COOKIE);
+		if (opt == NULL) {
+			ERR("bad cookie, missing EDNS section\n");
+			knot_pkt_free(reply);
+			return -1;
+		}
+
+		const uint8_t *data = knot_edns_opt_get_data(opt);
+		uint16_t data_len = knot_edns_opt_get_length(opt);
+		int ret = knot_edns_cookie_parse(&new_ctx.cc, &new_ctx.sc,
+		                                 data, data_len);
+		if (ret != KNOT_EOK) {
+			knot_pkt_free(reply);
+			ERR("bad cookie, missing EDNS cookie option\n");
+			return -1;
+		}
+		knot_pkt_free(reply);
+
+		// Restore the original client cookie.
+		new_ctx.cc = query_ctx->cc;
+
+		knot_pkt_t *new_query = create_query_packet(&new_ctx);
+		ret = process_query_packet(new_query, net, &new_ctx, ignore_tc,
+		                           sign_ctx, style);
+		knot_pkt_free(new_query);
+
+		return ret;
+	}
+
+	knot_pkt_free(reply);
 	net_close(net);
 
 	return 0;
@@ -777,7 +842,7 @@ static void process_query(const query_t *query)
 			if (ret == 0) {
 				net_clean(&net);
 				sign_context_deinit(&sign_ctx);
-				knot_pkt_free(&out_packet);
+				knot_pkt_free(out_packet);
 				return;
 			}
 
@@ -801,7 +866,7 @@ static void process_query(const query_t *query)
 	}
 
 	sign_context_deinit(&sign_ctx);
-	knot_pkt_free(&out_packet);
+	knot_pkt_free(out_packet);
 }
 
 static int process_xfr_packet(const knot_pkt_t      *query,
@@ -861,7 +926,7 @@ static int process_xfr_packet(const knot_pkt_t      *query,
 			} else {
 				ERR("can't print query packet\n");
 			}
-			knot_pkt_free(&q);
+			knot_pkt_free(q);
 		} else {
 			ERR("can't print query packet\n");
 		}
@@ -900,7 +965,7 @@ static int process_xfr_packet(const knot_pkt_t      *query,
 		// Parse reply to the packet structure.
 		if (knot_pkt_parse(reply, 0) != KNOT_EOK) {
 			ERR("malformed reply packet from %s\n", net->remote_str);
-			knot_pkt_free(&reply);
+			knot_pkt_free(reply);
 			net_close(net);
 			return 0;
 		}
@@ -908,7 +973,7 @@ static int process_xfr_packet(const knot_pkt_t      *query,
 		// Compare reply header id.
 		if (check_reply_id(reply, query) == false) {
 			ERR("reply ID mismatch from %s\n", net->remote_str);
-			knot_pkt_free(&reply);
+			knot_pkt_free(reply);
 			net_close(net);
 			return 0;
 		}
@@ -922,7 +987,7 @@ static int process_xfr_packet(const knot_pkt_t      *query,
 		if (knot_pkt_ext_rcode(reply) != KNOT_RCODE_NOERROR) {
 			ERR("server replied with error '%s'\n",
 			    knot_pkt_ext_rcode_name(reply));
-			knot_pkt_free(&reply);
+			knot_pkt_free(reply);
 			net_close(net);
 			return 0;
 		}
@@ -942,7 +1007,7 @@ static int process_xfr_packet(const knot_pkt_t      *query,
 
 					ERR("reply verification for %s (%s)\n",
 					    net->remote_str, knot_strerror(ret));
-					knot_pkt_free(&reply);
+					knot_pkt_free(reply);
 					net_close(net);
 					return 0;
 				}
@@ -954,7 +1019,7 @@ static int process_xfr_packet(const knot_pkt_t      *query,
 			if (serial < 0) {
 				ERR("first answer record from %s isn't SOA\n",
 				    net->remote_str);
-				knot_pkt_free(&reply);
+				knot_pkt_free(reply);
 				net_close(net);
 				return 0;
 			}
@@ -975,11 +1040,11 @@ static int process_xfr_packet(const knot_pkt_t      *query,
 
 		// Check for finished transfer.
 		if (finished_xfr(serial, reply, msg_count, query_ctx->serial != -1)) {
-			knot_pkt_free(&reply);
+			knot_pkt_free(reply);
 			break;
 		}
 
-		knot_pkt_free(&reply);
+		knot_pkt_free(reply);
 	}
 
 	// Get stop reply time.
@@ -1038,7 +1103,7 @@ static void process_xfr(const query_t *query)
 	               flags, &query->tls, &net);
 	if (ret != KNOT_EOK) {
 		sign_context_deinit(&sign_ctx);
-		knot_pkt_free(&out_packet);
+		knot_pkt_free(out_packet);
 		return;
 	}
 
@@ -1064,7 +1129,7 @@ static void process_xfr(const query_t *query)
 
 	net_clean(&net);
 	sign_context_deinit(&sign_ctx);
-	knot_pkt_free(&out_packet);
+	knot_pkt_free(out_packet);
 }
 
 int kdig_exec(const kdig_params_t *params)
